/**
 * [NOTE] The code of the tGrepper class isn't as streamlined (or logical) as
 *        it should have been.
 *
 * Right now, the GrepperThread() function calls GetNextName() every time it
 * wants to have the name of the next file to process. GetNextName() uses a
 * BList object that approaches a stack-like structure to remember the names 
 * of all subdirectories that it recurses into. A more sensible approach 
 * would be to move this functionality inside GrepperThread() and let it call 
 * itself recursively (this way we can use the implicit stackframe that C++ 
 * creates, instead of the clumsy BList code). Then again, maybe not.
 */
  
// ***************************************************************************
// Implementation System Includes
// ***************************************************************************

#include </boot/develop/headers/posix/sys/stat.h>       // struct stat

// ***************************************************************************
// Implementation Project Includes
// ***************************************************************************

#include "Grepper.h"

// ***************************************************************************
// Implementation Constant Definitions
// ***************************************************************************

/** Name of the temporary file that we will redirect grep's output to. */
//#define TEMPORARY_FILE "/boot/home/TrackerGrepResults"

// ***************************************************************************
// Implementation Type Definitions
// ***************************************************************************

// ***************************************************************************
// Implementation Variable Definitions
// ***************************************************************************

// ***************************************************************************
// Implementation Function Declarations
// ***************************************************************************

// ***************************************************************************
// Global Variable Definitions
// ***************************************************************************

// ***************************************************************************
// PUBLIC Member Function Definitions: Lifecycle
// ***************************************************************************

// ===========================================================================
// Constructor
// ===========================================================================

tGrepper::tGrepper(entry_ref aDirRef, BMessage* apMessage) 
{
  // Initialise attributes.
  mpMessage = new BMessage(apMessage);// _KMSF_012099 : make a copy...
  mpTarget = NULL;
//  mpPattern = NULL;
  SetPattern("");
  mRecurseDirs = TRUE;
  mRecurseLinks = FALSE;
  mThreadMustQuit = FALSE;
  mGrepperThreadId = -1;
  
  mpDirs = new BList(10);// _KMSF_012099_

  // Allocate a BDirectory object to remember the initial working directory.
  mpInitialDir = new BDirectory(&aDirRef);
  mpInitialDir->Rewind();
}

// ===========================================================================
// Destructor
// ===========================================================================

tGrepper::~tGrepper()
{
  // Wait for the grepper thread to quit.
  Stop();
  
  // Clean up.
  delete mpInitialDir;
  delete mpDirs;
  delete mpMessage;
}

// ***************************************************************************
// PUBLIC Member Function Definitions: Operators
// ***************************************************************************

// ***************************************************************************
// PUBLIC Member Function Definitions: Operations
// ***************************************************************************

// ===========================================================================
// Start
// ===========================================================================

status_t
tGrepper::Start()
{
//  // Do not continue if search pattern is empty.
//  if (NULL == mpPattern) {
//    return B_ERROR;
//  }

  // Create a new BList to hold the subdirectory objects and add _a_copy_of_ 
  // the initial directory to it. (We need to add a _copy_ to preserve the 
  // original initial directory object, since we will delete the entire list 
  // and its contents in the Stop() function.)
  //mpDirs = new BList(10);// _KMSF_012099 : init in constructor
  mpCurrentDir = new BDirectory(*mpInitialDir);
  mpDirs->AddItem((tVoid*) mpCurrentDir);
  mCurrentRef = 0;
  
  // Start the new thread.
  puts("[Tracker Grep] coucou1");
  mThreadMustQuit = FALSE;
  mGrepperThreadId = spawn_thread(SpawnThread, 
                                  "GrepperThread", 
//                                  B_LOW_PRIORITY,
                                  B_NORMAL_PRIORITY,
                                  this); 
                                  
  puts("[Tracker Grep] coucou2");
  return (resume_thread(mGrepperThreadId)); 
}

// ===========================================================================
// Stop
// ===========================================================================

tVoid
tGrepper::Stop()
{
  // _KMSF_012099_
  /* 
  		we already know the thread ID,
  		so just wait for it and if it's already dead,
  		wait_for_thread only fails...
  		and if someone launch anothre TG, it's not safe to ask for "GrepperThread"...
  */
  // Figure out if the grepper thread is actually running.
  //mGrepperThreadId = find_thread("GrepperThread");
  //if (B_NAME_NOT_FOUND != mGrepperThreadId) {

	puts("[Tracker Grep] stop");
    // Tell the thread to quit, and wait until it actually has.
    mThreadMustQuit = TRUE;
    tInt32 lExitValue;
    wait_for_thread(mGrepperThreadId, &lExitValue);

    // Free all BDirectory objects. (If the thread terminated normally, then
    // there is only one object: the copy of the initial directory. But if 
    // the user aborted the search, there may be more objects in the list.)
    // _KMSF_012099_
    // : delete AND remove BDirectory
    for (tInt32 count=mpDirs->CountItems();count>0;--count) {
    	delete (mpDirs->RemoveItem((tInt32)0));
    }
    /*
    tInt32 lIndex = 0;
    tVoid* lpItem = mpDirs->ItemAt(lIndex); 
    while (NULL != lpItem) {
      delete lpItem; 
      ++lIndex;
      lpItem = mpDirs->ItemAt(lIndex);
    }
    */
  
  	// _KMSF_012099_ : don't...
    // Always delete the list, even if it is empty.
    //delete mpDirs; 
    //mpDirs = NULL;
  //}
}  
    
// ***************************************************************************
// PUBLIC Member Function Definitions: Access
// ***************************************************************************

// ***************************************************************************
// PUBLIC Member Function Definitions: Inquiry
// ***************************************************************************

// ***************************************************************************
// PROTECTED Member Function Definitions
// ***************************************************************************

// ***************************************************************************
// PRIVATE Member Function Definitions
// ***************************************************************************

// ===========================================================================
// SpawnThread
// ===========================================================================

int32 
tGrepper::SpawnThread(void* apArg) 
{ 
  tGrepper* lpObj = (tGrepper*) apArg; 
  return (lpObj->GrepperThread()); 
}
   
// ===========================================================================
// GrepperThread
// ===========================================================================

tInt32 
tGrepper::GrepperThread() 
{
  tChar lFileName[MAX_LEN_REPORT_TEXT]; 
  tChar lTempString[MAX_LEN_REPORT_TEXT];
  tChar lCommand[B_PATH_NAME_LENGTH+32];
  BMessage* lpMsg;
  
  puts("[Tracker Grep] running");
  // init temp file _KMSF_012099_
  BPath tempfile;
  //if (find_directory(B_COMMON_VAR_DIRECTORY,&tempfile,true)!=B_OK) return (-1);
  char filename[B_FILE_NAME_LENGTH+30];
  sprintf(filename,"/boot/var/tmp/temp%ld",mGrepperThreadId);
  //tempfile.Append("/tmp/");
  tempfile.SetTo(filename);
  printf("[Tracker Grep] ");
  puts(filename);
  printf("[Tracker Grep] ");
  puts(tempfile.Path());

  // Loop through all the available files.
  while (B_OK == GetNextName(lFileName)) {

    // Are we told to quit ? Then exit the loop (this terminates the thread).
    if (TRUE == mThreadMustQuit) {
      break;
    }  

    // Send the name of the file to the target.
    //strcpy(lTempString,"Search results for \"");
    //strcat(lTempString,lFileName);
    //strcat(lTempString,"\":\n");
    strcpy(lTempString,lFileName);
    lpMsg = new BMessage(MSG_REPORT_FILE_NAME);
    lpMsg->AddString("filename",lTempString);
    mpTarget->PostMessage(lpMsg);
    delete lpMsg;

    // If the file's contents actually match the pattern, we will send another
    // message that also contains the entry_ref of the file for further use...
    // (like launching the file when it's double-clicked)
    lpMsg = new BMessage(MSG_REPORT_RESULT);
    lpMsg->AddString("filename",lTempString);
    
    BEntry entry(lFileName);
    entry_ref ref;
    entry.GetRef(&ref);
    lpMsg->AddRef("ref",&ref);

    // Execute grep and redirect its output to a temporary file.    
    AdjustFunnyCharacters(lFileName);
    sprintf(lCommand,
            "/boot/beos/bin/grep -hin %s \"%s\" > \"%s\"", // modified by Peter Hinely
//            mpPattern,
            mPattern,
            lFileName,
            tempfile.Path());// _KMSF_012099_
    system(lCommand);

    // Open the temporary file and send its contents to the target.    
    FILE* lpResults = fopen(tempfile.Path(),"r"); 
    if (lpResults) {

      while (NULL != fgets(lTempString,
                           MAX_LEN_REPORT_TEXT,
                           lpResults)) {
                     
        // _KMSF_011999_
        //lpMsg = new BMessage(MSG_REPORT_RESULT);
        lpMsg->AddString("text",lTempString);
        //mpTarget->PostMessage(lpMsg);
        //delete lpMsg;
        
        // Are we told to quit ? Then exit the loop.
        if (TRUE == mThreadMustQuit) {
          break;
        }  
        
        // _KMSF_011999_
       	// : why snoozing, we're multi-threading ?
        // Give the rest of the world some time to breathe.
        //snooze(20000);
      }

      // This file is finished.
      if (lpMsg->HasString("text")) {
      	mpTarget->PostMessage(lpMsg);
      }
      delete lpMsg;
      fclose(lpResults);
      if (FALSE == mThreadMustQuit) {
      // _KMSF_011999_!!!
      // should do something more clever with error
/*
        lpMsg = new BMessage(MSG_REPORT_RESULT);
        lpMsg->AddString("text","\n\n");
        mpTarget->PostMessage(lpMsg);
        delete lpMsg;     
*/ 
      }
    
    // There was a problem running grep.
      // _KMSF_011999_!!!
      // should do something more clever with error
    } else {
      lpMsg = new BMessage(MSG_REPORT_RESULT);
      lpMsg->AddString("text","There was a problem running grep.\n\n\n");
      mpTarget->PostMessage(lpMsg);
      delete lpMsg;      
    }
  }

  // We remove the temporary file after the _entire_ search has finished, to
  // prevent a whole lot of flickering in the corresponding Tracker window.
  remove(tempfile.Path());

  // Tell the target that the search has ended, unless we were terminated.
  if (FALSE == mThreadMustQuit) {
    BMessage lMsg(MSG_SEARCH_FINISHED);
    mpTarget->PostMessage(&lMsg);  
  }   
  return 0;
}

// ===========================================================================
// GetNextName
// ===========================================================================

status_t 
tGrepper::GetNextName(tChar* apBuffer)
{
  // We will loop until we have found the name of a valid file, or until 
  // there are no more files left.
  while (TRUE) {
  
    // Get the BEntry of the next name. --------------------------------------

    *apBuffer = NULL;                // Remove the previous string
    BEntry lEntry;
  
    // We are not inside a subdirectory.
    tInt32 lDirDepth = mpDirs->CountItems();
    if (1 == lDirDepth) { 

      // There are two ways to retrieve the file names. If the user selected
      // one or more files, then we must look at the message that was passed
      // into tGrepper's constructor. If the user didn't select any files, we
      // will simply read all the entries from the current working directory.
          
      // We have found an entry_ref in the message.
      entry_ref lFileRef;
      if (B_NO_ERROR == mpMessage->FindRef("refs",mCurrentRef,&lFileRef)) {

        // Go to next entry_ref in the message.
        ++mCurrentRef;

        // Convert the entry_ref into a BEntry.
        lEntry.SetTo(&lFileRef, mRecurseLinks);

      // If there were no files selected, we will grep the entire directory.
      } else if (0 == mCurrentRef) {

        // Get next entry from the current working directory.  
        if (B_OK != mpCurrentDir->GetNextEntry(&lEntry, mRecurseLinks)) {
          return B_ERROR;
        }

      // There are no more entry_refs in the message.
      } else { // mCurrentRef > 0
        return B_ERROR;
      }
      
    // We are inside a subdirectory.
    } else {

      // Get next entry from the current working directory.  
      if (B_OK != mpCurrentDir->GetNextEntry(&lEntry, mRecurseLinks)) {

        // No more entries in this subdirectory, so return to its parent.
        mpDirs->RemoveItem(mpCurrentDir);
        delete mpCurrentDir;
        mpCurrentDir = (BDirectory*) mpDirs->LastItem();

        // Immediately start a new iteration of the loop (to read the next 
        // BEntry from the parent subdirectory).
        continue;
      }
    }

    // Examine entry. --------------------------------------------------------

    // Get stat structure to determine whether the entry is a file or subdir.
    struct stat lStat;
    if (B_OK == lEntry.GetStat(&lStat)) { 
    
      // The entry is a directory.
      if (S_ISDIR(lStat.st_mode)) {

        // Do we want to recurse subdirectories ?
        BDirectory* lpDir = new BDirectory(&lEntry);
        if ((B_OK == lpDir->InitCheck()) && (TRUE == mRecurseDirs)) {
     
          mpDirs->AddItem((tVoid*) lpDir);  // Add new directory to the list
          mpCurrentDir = lpDir;             // and make it the current one

        // We won't use this directory object any more, so destroy it.
        } else {
          delete lpDir;
        }
              
      // The entry is a file or a (non-traversed) symbolic link.
      } else {
      
        // Because grep was meant to look at text files, we will add only 
        // those files to the list that have the MIME supertype "text" or 
        // "message". (Note that it is possible for grep to look at binary 
        // files, but the output is usually very ugly.) At this point the 
        // BEntry object of a symbolic link must already have been traversed
        // because the MIME type of a symbolic link itself is 
        // "application/x-vnd.Be-symlink" (which will be skipped).
    
        // Get the MIME type of the file.
        BNode lNode(&lEntry);
        BNodeInfo lNodeInfo(&lNode);
        tChar lMimeTypeString[B_MIME_TYPE_LENGTH];

        if (B_NO_ERROR == lNodeInfo.GetType(lMimeTypeString)) {

          // Only continue if the MIME supertype is "text" or "message".
          BMimeType lMimeType(lMimeTypeString);
          BMimeType lSuperType;         
          if (B_NO_ERROR == lMimeType.GetSupertype(&lSuperType)) {
            if ((0 == strcmp("text",lSuperType.Type())) 
            ||  (0 == strcmp("message",lSuperType.Type()))) {

              // Get the full path name and copy it into the buffer.
              BPath lPath;
              if (B_OK == lEntry.GetPath(&lPath)) { 
                strcpy(apBuffer,lPath.Path());
                return B_OK;
              }
            } 
          }
        } // if GetType()
      } // if entry is file or symbolic link
    } // if GetStat()

    // We did find a file or directory but it wasn't valid for our purposes.
    // So we'll just try again.
    // snooze(20000);
  }
}

// ===========================================================================
// AdjustFunnyCharacters
// ===========================================================================

tVoid
tGrepper::AdjustFunnyCharacters(tChar* apBuffer) 
{
  tChar* lpCopyOfBuffer = strdup(apBuffer);
  for (tUInt32 lCount=0; lCount<strlen(lpCopyOfBuffer); ++lCount) {
    if ('"' == lpCopyOfBuffer[lCount]) {
      *apBuffer = '\\';      
      ++apBuffer;
      *apBuffer = '"';
    } else {
      *apBuffer = lpCopyOfBuffer[lCount];
    }
    ++apBuffer;
  }
  *apBuffer='\0';
  free(lpCopyOfBuffer);
}

// ***************************************************************************
// Helper Function Definitions
// ***************************************************************************

// ***************************************************************************
// Global Function Definitions
// ***************************************************************************

// ***************************************************************************
// Implementation Function Definitions
// ***************************************************************************
